#include <kernel/fs.h>
#include <errno.h>
#include <kernel/proc.h>
#include <libgen.h>
#include <kernel/k_libc.h>
//----------------------------------------------------------------------
int
path_mkdir (pid_t pid, const char *path, mode_t mode)
{
    proc_t   *ps;
    inode_t  *inode_directory;
    inode_t  *inode_parent;
    int       status;
    char      path_directory[PATH_MAX];
    char      path_copy[PATH_MAX];
    char     *path_parent;
    ssize_t   size_written;
    //
    struct {
        ino_t inode_1;
        char  name_1[NAME_MAX];
        ino_t inode_2;
        char  name_2[NAME_MAX];
    } directory;
    //
    // Get process.
    //
    ps = proc_reference (pid);
    //
    // Correct the mode with the umask.
    //
    mode &= ~ps->umask;
    //
    // Inside `mode', the file type is fixed. No check is made.
    //
    mode &= 00777;
    mode |= S_IFDIR;
    //
    // The full path and the directory path is needed.
    //
    status = path_full (path, ps->path_cwd, path_directory);
    if (status < 0)
      {
        return (-1);
      }
    strncpy (path_copy, path_directory, PATH_MAX);
    path_copy[PATH_MAX-1] = 0;
    path_parent = dirname (path_copy);
    //
    // Check if something already exists with the same name. The scan
    // is done with kernel privileges.
    //
    inode_directory = path_inode ((uid_t) 0, path_directory);
    if (inode_directory != NULL)
      {
        //
        // The file already exists. Put inode and return an error.
        //
        inode_put (inode_directory);
        errset (EEXIST);                // File exists.
        return (-1);
      } 
    //
    // Try to locate the directory that should contain this one.
    //
    inode_parent = path_inode (pid, path_parent);
    if (inode_parent == NULL)
      {
        //
        // Cannot locate the directory: return an error. The variable
        // `errno' should already be set by `path_inode()'.
        //
        errset (errno);
        return (-1);
      }
    //
    // Try to create the node: should fail if the user does not have
    // enough permissions.
    //
    inode_directory = path_inode_link (pid, path_directory, NULL,
                                       mode);
    if (inode_directory == NULL)
      {
        //
        // Sorry: cannot create the inode! The variable `errno' should
        // already be set by `path_inode_link()'.
        //
        errset (errno);
        return (-1);
      }
    //
    // Fill records for `.' and `..'.
    //
    directory.inode_1 = inode_directory->ino;
    strncpy (directory.name_1, ".",  (size_t) 3);
    directory.inode_2 = inode_parent->ino;
    strncpy (directory.name_2, "..", (size_t) 3);
    //
    // Write data.
    //
    size_written = inode_file_write (inode_directory, (off_t) 0,
                                     &directory, (sizeof directory));
    if (size_written != (sizeof directory))
      {
        return (-1);
      }
    //
    // Fix directory inode links.
    //
    inode_directory->links   = 2;
    inode_directory->time    = k_time (NULL);
    inode_directory->changed = 1;
    //
    // Fix parent directory inode links.
    //
    inode_parent->links++;
    inode_parent->time       = k_time (NULL);
    inode_parent->changed    = 1;
    //
    // Save and put the inodes.
    //
    inode_save (inode_parent);
    inode_save (inode_directory);
    inode_put  (inode_parent);
    inode_put  (inode_directory);
    //
    // Return.
    //
    return (0);
}
