#include <kernel/proc.h>
#include <kernel/k_libc.h>
//----------------------------------------------------------------------
void
proc_sch_terminals (void)
{
    pid_t  pid;
    int    key;
    tty_t *tty;
    dev_t  device;
    //
    // Try to read a key from console keyboard buffer (only consoles
    // are available).
    //
    key = con_char_ready ();
    if (key == 0)
      {
        //
        // No key is ready on the keyboard buffer: just return.
        //
        return;
      }
    //
    // A key is available. Find the currently active console.
    //
    device = tty_console ((dev_t) 0);
    tty    = tty_reference (device);
    if (tty == NULL)
      {
        k_printf ("kernel alert: console device 0x%04x not found!\n",
                  device);
        //
        // Will send the typed character to the first terminal!
        //
        tty = tty_reference ((dev_t) 0);
      }
    //
    // Defined the active console. Put the character there.
    //
    if (tty->key == 0)
      {
        tty->status = TTY_OK;
      }
    else
      {
        tty->status = TTY_LOST_KEY;
      }
    tty->key = con_char_read ();
    //
    // Verify if it is a control key that must be handled. If so, a
    // signal is sent to all processes with the same control terminal,
    // excluded the kernel (0) and `init' (1). Such control keys are not
    // passed to the applications.
    //
    // Please note that this a simplified solution, because the signal
    // should reach only the foreground process of the group. For that
    // reason, only che [Ctrl C] is taken into consideration, because
    // processes can ignore the signal 'SIGINT'.
    //
    if (tty->pgrp != 0)
      {
        //
        // There is a process group for that terminal.
        //
        if      (tty->key == 3)         // [Ctrl C] -> SIGINT
          {
            for (pid = 2; pid <  PROCESS_MAX; pid++)
              {
                if (proc_table[pid].pgrp == tty->pgrp)
                  {
                    k_kill (pid, SIGINT);
                  }
              }
            tty->key    = 0;    // Reset key and status.
            tty->status = TTY_OK;
          }
      }
    //
    // Check for a console switch key combination.
    //
    if      (tty->key == 0x11)          // [Ctrl Q] -> DC1 -> console0.
      {
        tty->key    = 0;                // Reset key and status.
        tty->status = TTY_OK;
        tty_console (DEV_CONSOLE0);     // Switch.
      }
    else if (tty->key == 0x12)          // [Ctrl R] -> DC2 -> console1.
      {
        tty->key    = 0;                // Reset key and status.
        tty->status = TTY_OK;
        tty_console (DEV_CONSOLE1);     // Switch.
      }
    else if (tty->key == 0x13)          // [Ctrl S] -> DC3 -> console2.
      {
        tty->key    = 0;                // Reset key and status.
        tty->status = TTY_OK;
        tty_console (DEV_CONSOLE2);     // Switch.
      }
    else if (tty->key == 0x14)          // [Ctrl T] -> DC4 -> console3.
      {
        tty->key    = 0;                // Reset key and status.
        tty->status = TTY_OK;
        tty_console (DEV_CONSOLE3);     // Switch.
      }
    //
    // A key was pressed: must wake up all processes waiting for reading
    // a terminal: all processes must be reactivated, because a process
    // can read from the device file, and not just from its own
    // terminal.
    //
    for (pid = 0; pid < PROCESS_MAX; pid++)
      {
        if (   (proc_table[pid].status == PROC_SLEEPING)
            && (proc_table[pid].wakeup_events & WAKEUP_EVENT_TTY))
          {
            //
            // A process waiting for that terminal was found:
            // remove the waiting event and set it ready.
            //
            proc_table[pid].wakeup_events &= ~WAKEUP_EVENT_TTY;
            proc_table[pid].status = PROC_READY;
          }
      }
}
