#include <kernel/proc.h>
#include <errno.h>
#include <fcntl.h>
//----------------------------------------------------------------------
int
proc_sys_exec (uint16_t *sp, segment_t *segment_d, pid_t pid,
               const char *path,
               unsigned int argc, char *arg_data,
               unsigned int envc, char *env_data)
{
    unsigned int      i;
    unsigned int      j;
    char             *arg;
    char             *env;
    char             *envp[ARG_MAX/16];
    char             *argv[ARG_MAX/16];
    size_t            size;
    size_t            arg_data_size;
    size_t            env_data_size;
    unsigned int      p_off;
    inode_t          *inode;
    ssize_t           size_read;
    header_t          header;
    unsigned long int s;        // Help calculating process sizes.
    uint16_t          new_sp;
    uint16_t          envp_address;
    uint16_t          argv_address;
    size_t            process_size_i;
    size_t            process_size_d;
    char              buffer[MEM_BLOCK_SIZE];
    uint16_t          stack_element;
    off_t             inode_start;
    addr_t            memory_start;
    addr_t            previous_address_i;
    segment_t         previous_segment_i;
    size_t            previous_size_i;
    addr_t            previous_address_d;
    segment_t         previous_segment_d;
    size_t            previous_size_d;
    int               status;
    memory_t          allocated_i;
    memory_t          allocated_d;
    pid_t             extra;
    int               proc_count;
    file_t           *file;
    int               fdn;
    dev_t             device;
    int               eof;
    //
    // Check for limits.
    //
    if (argc > (ARG_MAX/16) || envc > (ARG_MAX/16))
      {
        errset (ENOMEM);
        return (-1);
      }
    //
    // Scan arguments to calculate the full size and the relative
    // pointers. The final size is rounded to 2, for the stack.
    //
    arg = arg_data;
    for (i = 0, j = 0; i < argc; i++)
      {
        argv[i] = (char *) j;   // Relative pointer inside the `arg_data'.
        size = strlen (arg);
        arg += size + 1;
        j += size + 1;
      }
    arg_data_size = j;
    if (arg_data_size % 2)
      {
        arg_data_size++;
      }
    //
    // Scan environment variables to calculate the full size and the
    // relative pointers. The final size is rounded to 2, for the stack.
    //
    env = env_data;
    for (i = 0, j = 0; i < envc; i++)
      {
        envp[i] = (char *) j;   // Relative pointer inside the `env_data'.
        size = strlen (env);
        env += size + 1;
        j += size + 1;
      }
    env_data_size = j;
    if (env_data_size % 2)
      {
        env_data_size++;
      }
    //
    // Read the inode related to the executable file name.
    // Function path_inode() includes the inode get procedure.
    //
    inode = path_inode (pid, path);
    if (inode == NULL)
      {
        errset (ENOENT);        // No such file or directory.
        return (-1);
      }
    //
    // Check for permissions.
    //
    status = inode_check (inode, S_IFREG, 5, proc_table[pid].euid);
    if (status != 0)
      {
        //
        // File is not of a valid type or permission are not
        // sufficient: release the executable file inode
        // and return with an error.
        //
        inode_put (inode);
        errset (EACCES);        // Permission denied.
        return (-1);
      }
    //
    // Read the header from the executable file.
    //
    size_read = inode_file_read (inode, (off_t) 0, &header,
                                 (sizeof header), &eof);
    if (size_read != (sizeof header))
      {
        //
        // The file is shorter than the executable header, so, it isn't
        // an executable: release the file inode and return with an
        // error.
        //
        inode_put (inode);
        errset (ENOEXEC);
        return (-1);
      }
    if (header.magic0 != MAGIC_OS16 || header.magic1 != MAGIC_OS16_APPL)
      {
        //
        // The header does not have the expected magic numbers, so,
        // it isn't a valid executable: release the file inode and
        // return with an error.
        //
        inode_put (inode);
        errset (ENOEXEC);
        return (-1);    // This is not a valid executable!
      }
    //
    // Calculate data size.
    //
    s  = header.ebss;
    if (header.ssize == 0)
      {
        s += 0x10000L;  // Zero means max size.
      }
    else
      {
        s += header.ssize;
      }
    if (s > 0xFFFF)
      {
        process_size_d = 0x0000;  // 0x0000 means the maximum size:
                                  // 0x10000.
        new_sp         = 0x0000;  // 0x0000 is like 0x10000 and the first
                                  // push moves SP to 0xFFFE.
      }
    else
      {
        process_size_d = s;
        new_sp         = process_size_d;
        if (new_sp % 2)
          {
            new_sp--;       // The stack pointer should be even.
          }
      }
    //
    // Calculate code size.
    //
    if (header.segoff == 0)
      {
        process_size_i = process_size_d;
      }
    else
      {
        process_size_i = header.segoff * 16;
      }
    //
    // Allocate memory: code and data segments.
    //
    status = mb_alloc_size (process_size_i, &allocated_i);
    if (status < 0)
      {
        //
        // The program instructions (code segment) cannot be loaded
        // into memory: release the executable file inode and return
        // with an error.
        //
        inode_put (inode);
        errset (ENOMEM);        // Not enough space.
        return ((pid_t) -1);
      }
    if (header.segoff == 0)
      {
        //
        // Code and data segments are the same: no need
        // to allocate more memory for the data segment.
        //
        allocated_d.address = allocated_i.address;
        allocated_d.segment = allocated_i.segment;
        allocated_d.size    = allocated_i.size;
      }
    else
      {
        //
        // Code and data segments are different: the data
        // segment memory is allocated.
        //
        status = mb_alloc_size (process_size_d, &allocated_d);
        if (status < 0)
          {
            //
            // The separated program data (data segment) cannot be loaded
            // into memory: free the already allocated memory for the
            // program instructions, release the executable file inode
            // and return with an error.
            //
            mb_free (allocated_i.address, allocated_i.size);
            inode_put (inode);
            errset (ENOMEM);        // Not enough space.
            return ((pid_t) -1);
          }
      }
    //
    // Load executable in memory.
    //
    if (header.segoff == 0)
      {
        //
        // Code and data share the same segment.
        //
        for (eof = 0, memory_start = allocated_i.address,
             inode_start = 0, size_read = 0;
             inode_start < inode->size && !eof;
             inode_start += size_read)
          {
            memory_start += size_read;
            //
            // Read a block of memory.
            //
            size_read = inode_file_read (inode, inode_start,
                                         buffer, MEM_BLOCK_SIZE, &eof);
            if (size_read < 0)
              {
                //
                // Free memory and inode.
                //
                mb_free (allocated_i.address, allocated_i.size);
                inode_put (inode);
                errset (EIO);
                return (-1);
              }
            //
            // Copy inside the right position to be executed.
            //
            dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE, memory_start, buffer,
                    (size_t) size_read, NULL);
          }
      }
    else
      {
        //
        // Code and data with different segments.
        //
        for (eof = 0, memory_start = allocated_i.address,
             inode_start = 0, size_read = 0;
             inode_start < process_size_i && !eof;
             inode_start += size_read)
          {
            memory_start += size_read;
            //
            // Read a block of memory
            //
            size_read = inode_file_read (inode, inode_start,
                                         buffer, MEM_BLOCK_SIZE, &eof);
            if (size_read < 0)
              {
                //
                // Free memory and inode.
                //
                mb_free (allocated_i.address, allocated_i.size);
                mb_free (allocated_d.address, allocated_d.size);
                inode_put (inode);
                errset (EIO);
                return (-1);
              }
            //
            // Copy inside the right position to be executed.
            //
            dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE, memory_start, buffer,
                    (size_t) size_read, NULL);
          }
        for (eof = 0, memory_start = allocated_d.address,
             inode_start = (header.segoff * 16), size_read = 0;
             inode_start < inode->size && !eof;
             inode_start += size_read)
          {
            memory_start += size_read;
            //
            // Read a block of memory
            //
            size_read = inode_file_read (inode, inode_start,
                                         buffer, MEM_BLOCK_SIZE, &eof);
            if (size_read < 0)
              {
                //
                // Free memory and inode.
                //
                mb_free (allocated_i.address, allocated_i.size);
                mb_free (allocated_d.address, allocated_d.size);
                inode_put (inode);
                errset (EIO);
                return (-1);
              }
            dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE, memory_start, buffer,
                    (size_t) size_read, NULL);
          }
      }
    //
    // The executable file was successfully loaded in memory:
    // release the executable file inode.
    //
    inode_put (inode);
    //
    // Put environment data inside the stack.
    //
    new_sp -= env_data_size; //----------------------------- environment
    mem_copy (address (seg_d (), (unsigned int) env_data),
              (allocated_d.address + new_sp), env_data_size);
    //
    // Put arguments data inside the stack.
    //
    new_sp -= arg_data_size; //------------------------------- arguments
    mem_copy (address (seg_d (), (unsigned int) arg_data),
              (allocated_d.address + new_sp), arg_data_size);
    //
    // Put envp[] inside the stack, updating all the pointers.
    //
    new_sp -= 2; //------------------------------------------------ NULL
    stack_element = NULL;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    //
    p_off  = new_sp;            //
    p_off += 2;                 // Calculate memory pointers from
    p_off += arg_data_size;     // original relative pointers,
    for (i = 0; i < envc; i++)  // inside the environment array
      {                         // of pointers.
        envp[i] += p_off;       //
      }                         //
    //
    new_sp -= (envc * (sizeof (char *))); //-------------------- *envp[]
    mem_copy (address (seg_d (), (unsigned int) envp),
              (allocated_d.address + new_sp),
              (envc * (sizeof (char *))));
    //
    // Save the envp[] location, needed in the following.
    //
    envp_address = new_sp;
    //
    // Put argv[] inside the stack, updating all the pointers.
    //
    new_sp -= 2; //------------------------------------------------ NULL
    stack_element = NULL;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    //
    p_off  = new_sp;                      //
    p_off += 2 ;                          // Calculate memory pointers
    p_off += (envc * (sizeof (char *)));  // from original relative
    p_off += 2;                           // pointers, inside the
    for (i = 0; i < argc; i++)            // arguments array of
      {                                   // pointers.
        argv[i] += p_off;                 //
      }                                   //
    //
    new_sp -= (argc * (sizeof (char *))); //-------------------- *argv[]
    mem_copy (address (seg_d (), (unsigned int) argv),
              (allocated_d.address + new_sp),
              (argc * (sizeof (char *))));
    //
    // Save the argv[] location, needed in the following.
    //
    argv_address = new_sp;
    //
    // Put the pointer to the array envp[].
    //
    new_sp -= 2; //------------------------------------------------ argc
    stack_element = envp_address;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    //
    // Put the pointer to the array argv[].
    //
    new_sp -= 2; //------------------------------------------------ argc
    stack_element = argv_address;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    //
    // Put argc inside the stack.
    //
    new_sp -= 2; //------------------------------------------------ argc
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &argc, (sizeof argc), NULL);
    //
    // Set the rest of the stack.
    //
    new_sp -= 2; //----------------------------------------------- FLAGS
    stack_element = 0x0200;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- CS
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &allocated_i.segment, (sizeof allocated_i.segment), NULL);
    new_sp -= 2; //-------------------------------------------------- IP
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- ES
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &allocated_d.segment, (sizeof allocated_d.segment), NULL);
    new_sp -= 2; //-------------------------------------------------- DS
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &allocated_d.segment, (sizeof allocated_d.segment), NULL);
    new_sp -= 2; //-------------------------------------------------- DI
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- SI
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- BP
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- BX
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- DX
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- CX
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    new_sp -= 2; //-------------------------------------------------- AX
    stack_element = 0;
    dev_io ((pid_t) 0, DEV_MEM, DEV_WRITE,
            (allocated_d.address + new_sp),
            &stack_element, (sizeof stack_element), NULL);
    //
    // Close process file descriptors, if the `FD_CLOEXEC' flag
    // is present.
    //
    for (fdn = 0; fdn < OPEN_MAX; fdn++)
      {
        if (proc_table[pid].fd[0].file != NULL)
          {
            if (proc_table[pid].fd[0].fd_flags & FD_CLOEXEC)
              {
                fd_close (pid, fdn);
              }
          }
      }
    //
    // Select device for standard I/O, if a standard I/O stream must be
    // opened.
    //
    if (proc_table[pid].device_tty != 0)
      {
        device = proc_table[pid].device_tty;
      }
    else
      {
        device = DEV_TTY;
      }
    //
    // Prepare missing standard file descriptors. The function
    // `file_stdio_dev_make()' arranges the value for `errno' if
    // necessary. If a standard file descriptor cannot be allocated,
    // the program is left without it.
    //
    if (proc_table[pid].fd[0].file == NULL)
      {
        file = file_stdio_dev_make (device, S_IFCHR, O_RDONLY);
        if (file != NULL)                                      // stdin
          {
            proc_table[pid].fd[0].fl_flags     = O_RDONLY;
            proc_table[pid].fd[0].fd_flags     = 0;
            proc_table[pid].fd[0].file         = file;
            proc_table[pid].fd[0].file->offset = 0;
          }
      }
    if (proc_table[pid].fd[1].file == NULL)
      {
        file = file_stdio_dev_make (device, S_IFCHR, O_WRONLY);
        if (file != NULL)                                     // stdout
          {
            proc_table[pid].fd[1].fl_flags     = O_WRONLY;
            proc_table[pid].fd[1].fd_flags     = 0;
            proc_table[pid].fd[1].file         = file;
            proc_table[pid].fd[1].file->offset = 0;
          }
      }
    if (proc_table[pid].fd[2].file == NULL)
      {
        file = file_stdio_dev_make (device, S_IFCHR, O_WRONLY);
        if (file != NULL)                                     // stderr
          {
            proc_table[pid].fd[2].fl_flags     = O_WRONLY;
            proc_table[pid].fd[2].fd_flags     = 0;
            proc_table[pid].fd[2].file         = file;
            proc_table[pid].fd[2].file->offset = 0;
          }
      }
    //
    // Prepare to switch
    //
    previous_address_i = proc_table[pid].address_i;
    previous_segment_i = proc_table[pid].segment_i;
    previous_size_i    = proc_table[pid].size_i;
    previous_address_d = proc_table[pid].address_d;
    previous_segment_d = proc_table[pid].segment_d;
    previous_size_d    = proc_table[pid].size_d;
    //
    proc_table[pid].address_i  = allocated_i.address;
    proc_table[pid].segment_i  = allocated_i.segment;
    proc_table[pid].size_i     = allocated_i.size;
    proc_table[pid].address_d  = allocated_d.address;
    proc_table[pid].segment_d  = allocated_d.segment;
    proc_table[pid].size_d     = allocated_d.size;
    proc_table[pid].sp         = new_sp;
    strncpy (proc_table[pid].name, path, PATH_MAX);
    //
    // Ensure to have a terminated string.
    //
    proc_table[pid].name[PATH_MAX-1] = 0;
    //
    // Free data segment memory.
    //
    mb_free (previous_address_d, previous_size_d);
    //
    // Free code segment memory if it is
    // different from the data segment.
    //
    if (previous_segment_i != previous_segment_d)
      {
        //
        // Must verify if no other process is
        // using the same memory.
        //
        for (proc_count = 0, extra = 0; extra < PROCESS_MAX; extra++)
          {
            if (proc_table[extra].status == PROC_EMPTY   ||
                proc_table[extra].status == PROC_ZOMBIE)
              {
                continue;
              }
            if (previous_segment_i == proc_table[extra].segment_i)
              {
                proc_count++;
              }
          }
        if (proc_count == 0)
          {
            //
            // The code segment can be released, because no other
            // process is using it.
            //
            mb_free (previous_address_i, previous_size_i);
          }
      }
    //
    // Change the segment and the stack pointer, from the interrupt.
    //
    *segment_d = proc_table[pid].segment_d;
    *sp        = proc_table[pid].sp;
    //
    return (0);
}
